import math
from typing import Tuple, Union

from ..event import MaiNote, NoteType, Event, EventType

# Dictionary for a note type's representation in ma2
# Does not cover slide notes, BPM, and meter events.
# Use slide_dict for slides instead.
note_dict = {
    "TAP": 1,
    "HLD": 2,
    "BRK": 3,
    "STR": 4,
    "BST": 5,
    "XTP": 6,
    "XST": 7,
    "XHO": 8,
    "TTP": 9,
    "THO": 10,
    # "SLD": 11
}

# Dictionary for a slide note's representation in ma2 by pattern.
slide_dict = {
    "SI_": 1,
    "SCL": 2,
    "SCR": 3,
    "SUL": 4,
    "SUR": 5,
    "SSL": 6,
    "SSR": 7,
    "SV_": 8,
    "SXL": 9,
    "SXR": 10,
    "SLL": 11,
    "SLR": 12,
    "SF_": 13,
}


class SlideNote(MaiNote):
    def __init__(
        self,
        measure: float,
        start_position: int,
        end_position: int,
        pattern: int,
        duration: float,
        delay: float = 0.25,
    ) -> None:
        """Produces a ma2 slide note.

        Note:
            Please use MaiMa2 class' add_slide method for adding slides.

        Args:
            measure: Time when the note starts, in terms of measures.
            start_position: Starting button.
            end_position: Ending button.
            pattern: Numerical representation of the slide pattern.
            duration: Time duration of when the slide starts moving and
                      when it ends (delay is not included.) in terms of
                      measures.
            delay: Time duration of when the slide appears and when it
                   starts to move, in terms of measures.
            resolution: Ma2 time resolution used. Optional, defaults to
                        384.
        Raises:
            ValueError: When pattern or duration is not a positive integer,
                        and when delay, or end_position are negative.
        """
        if pattern <= 0:
            raise ValueError("Slide pattern is not positive " + str(pattern))
        if duration <= 0:
            raise ValueError("Slide duration is not positive " + str(duration))
        if delay < 0:
            raise ValueError("Slide delay is negative " + str(delay))
        if end_position < 0:
            raise ValueError("Slide end position is negative " + str(end_position))

        measure = round(10000.0 * measure) / 10000.0
        super().__init__(measure, start_position, NoteType.complete_slide)
        self.end_position = end_position
        self.pattern = pattern
        self.delay = delay
        self.duration = duration


class HoldNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        duration: float,
        is_ex: bool = False,
    ) -> None:
        """Produces a ma2 hold note.

        Note:
            Please use MaiMa2 class' add_hold method for adding holds.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the hold note happens.
            duration: Total time duration of the hold note.
            is_ex: Whether a hold note is an ex note.
            resolution: Ma2 time resolution used. Optional, defaults to
                        384.

        Raises:
            ValueError: When duration is not positive.
        """
        if duration < 0:
            raise ValueError(f"Hold duration is negative: {duration}")

        measure = round(measure * 10000.0) / 10000.0
        duration = round(duration * 10000.0) / 10000.0

        if is_ex:
            super().__init__(measure, position, NoteType.ex_hold)
        else:
            super().__init__(measure, position, NoteType.hold)

        self.duration = duration


class TapNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        is_star: bool = False,
        is_break: bool = False,
        is_ex: bool = False,
    ) -> None:
        """Produces a ma2 tap note.

        Note:
            Please use MaiMa2 class' add_tap method for adding taps.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the tap note happens.
            is_star: Whether a tap note is a star note.
            is_break: Whether a tap note is a break note.
            is_ex: Whether a tap note is an ex note.
            resolution: Ma2 time resolution used. Optional, defaults to
                        384.
        """
        measure = round(measure * 10000.0) / 10000.0

        if is_ex and is_star:
            super().__init__(measure, position, NoteType.ex_star)
        elif is_ex and not is_star:
            super().__init__(measure, position, NoteType.ex_tap)
        elif is_star and is_break:
            super().__init__(measure, position, NoteType.break_star)
        elif is_star and not is_break:
            super().__init__(measure, position, NoteType.star)
        elif not is_star and is_break:
            super().__init__(measure, position, NoteType.break_tap)
        elif not is_star and not is_break:
            super().__init__(measure, position, NoteType.tap)


class TouchTapNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        region: str,
        is_firework: bool = False,
        size: str = "M1",
    ) -> None:
        """Produces a ma2 touch tap note.

        Note:
            Please use MaiMa2 class' add_touch_tap method
            for adding touch taps.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Position in the region where the note happens.
            region: Touch region where the note happens.
            is_firework: Whether a touch tap note will produce
                         fireworks. Optional bool, defaults to False.
            size: Optional str. Specifies the size of touch note defaults to "M1"

        Examples:
            A touch tap note at E0 at measure 2.25, produces no
            fireworks.

            >>> touch_tap = MaiMa2TouchTapNote(2.25, 0, "E")

            A touch tap note at B5 at measure 5.00, produces fireworks.

            >>> touch_tap = MaiMa2TouchTapNote(5.00, 5, "B", True)
        """
        if size not in ["M1", "L1"]:
            raise ValueError(f"Invalid size given: {size}")

        measure = round(measure * 10000.0) / 10000.0

        super().__init__(measure, position, NoteType.touch_tap)
        self.is_firework = is_firework
        self.region = region
        self.size = size


class TouchHoldNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        region: str,
        duration: float,
        is_firework: bool = False,
        size: str = "M1",
    ) -> None:
        """Produces a ma2 touch hold note.

        Note:
            Please use MaiMa2 class' add_touch_hold method
            for adding touch holds.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Position in the region where the note happens.
            region: Touch region where the note happens.
            duration: Total time duration of the touch hold note.
            is_firework: Whether a touch hold note will produce
                         fireworks. Optional bool, defaults to False.
            size: Optional str. Specifies the size of touch note defaults to "M1"

        Examples:
            A touch hold note at C0 at measure 1 with duration of
            1.5 measures, produces no fireworks.

            >>> touch_tap = MaiMa2TouchHoldNote(1, 0, "C", 1.5)
        """
        if duration <= 0:
            raise ValueError(f"Hold duration is not positive: {duration} ")
        if size not in ["M1", "L1"]:
            raise ValueError(f"Invalid size given: {size}")

        measure = round(measure * 10000.0) / 10000.0
        duration = round(duration * 10000.0) / 10000.0

        super().__init__(measure, position, NoteType.touch_hold)
        self.duration = duration
        self.is_firework = is_firework
        self.region = region
        self.size = size


class BPM(Event):
    def __init__(self, measure: float, bpm: float) -> None:
        """Produces a ma2 BPM event.

        Note:
            Please use MaiMa2 class' set_bpm method
            for adding BPM events.

        Args:
            measure: Time when the BPM change happens,
                     in terms of measures.
            bpm: The tempo in beats per minute.
            resolution: Ma2 time resolution used. Optional, defaults to
                        384.

        Raises:
            ValueError: When bpm is not positive.

        Examples:
            A bpm of 220 at measure 3.

            >>> ma2_bpm = MaiMa2BPM(3, 220)
        """
        if bpm <= 0:
            raise ValueError(f"BPM is not positive: {bpm}")

        measure = round(measure * 10000.0) / 10000.0

        super().__init__(measure, EventType.bpm)
        self.bpm = bpm


class Meter(Event):
    def __init__(
        self,
        measure: float,
        meter_numerator: int,
        meter_denominator: int,
    ) -> None:
        """Produces a ma2 meter signature event.

        Note:
            Please use MaiMa2 class' set_meter method
            for adding meter signature events.

        Args:
            measure: Time when the meter change happens,
                     in terms of measures.
            meter_numerator: The upper numeral in a meter/time
                             signature.
            meter_denominator: The lower numeral in a meter/time
                             signature.
            resolution: Ma2 time resolution used. Optional, defaults to
                        384.

        Raises:
            ValueError: When meter_numerator or
                        meter_denominator is not positive.
        """
        if meter_numerator <= 0:
            raise ValueError("Meter numerator is not positive: {meter_numerator}")
        if meter_denominator <= 0:
            raise ValueError(f"Meter denominator is not positive {meter_denominator}")

        measure = round(measure * 10000.0) / 10000.0

        super().__init__(measure, EventType.meter)
        self.numerator = meter_numerator
        self.denominator = meter_denominator


def measure_to_ma2_time(measure: float, resolution: int = 384) -> Tuple[int, int]:
    """Convert measure in decimal form to ma2's format.

    Ma2 uses integer timing for its measures. It does so by taking
    the fractional part and multiplying it to the chart's
    resolution, then rounding it up. For durations like holds
    and slides, both the whole and fractional part are
    multiplied by the charts resolution.

    Args:
        measure: The time a note happened or duration.

    Returns:
        A tuple (WHOLE_PART, FRACTIONAL_PART).

    Raises:
        ValueError: When measure is negative.

    Examples:
        Convert a note that happens in measure 2.5 in ma2 time
        with resolution of 384.

        >>> print(measure_to_ma2_time(2.5))
        (2, 192)

        Convert a duration of 3.75 measures in ma2 time with
        resolution of 500.

        >>> my_resolution = 500
        >>> (whole, dec) = measure_to_ma2_time(3.75, my_resolution)
        >>> whole*my_resolution + dec
        1875.0
    """
    if measure < 0:
        raise ValueError("Measure is negative. " + str(measure))

    (decimal_part, whole_part) = math.modf(measure)
    decimal_part = round(decimal_part * resolution)

    return (int(whole_part), decimal_part)


def event_to_str(
    event: Union[TapNote, HoldNote, SlideNote, TouchTapNote, TouchHoldNote, BPM, Meter],
    resolution: int = 384,
) -> str:
    """Converts ma2 note and events into ma2-compatible lines.

    Args:
        event: A bpm, meter, or note event.

    Returns:
        A single line string with a line ending.

    Raises:
        TypeError: If a note or event is unknown.
        ValueError: If a slide note has unknown pattern.
    """
    if not isinstance(event, Event):
        raise TypeError("{} is not an Event type".format(event))

    inv_note_dict = {v: k for k, v in note_dict.items()}
    measure = measure_to_ma2_time(event.measure, resolution)

    if isinstance(event, BPM):  # BPM
        template = "BPM\t{}\t{}\t{:.3f}\n"
        result = template.format(measure[0], measure[1], event.bpm)
    elif isinstance(event, Meter):  # Meter
        template = "MET\t{}\t{}\t{}\t{}\n"
        result = template.format(
            measure[0], measure[1], event.numerator, event.denominator
        )
    elif isinstance(event, TapNote):
        template = "{}\t{}\t{}\t{}\n"
        name = inv_note_dict[event.note_type.value]
        result = template.format(name, measure[0], measure[1], event.position)
    elif isinstance(event, HoldNote):
        template = "{}\t{}\t{}\t{}\t{}\n"
        name = inv_note_dict[event.note_type.value]
        duration = round(event.duration * resolution)
        result = template.format(name, measure[0], measure[1], event.position, duration)
    elif isinstance(event, SlideNote):
        template = "{}\t{}\t{}\t{}\t{}\t{}\t{}\n"
        inv_slide_dict = {v: k for k, v in slide_dict.items()}
        if event.pattern not in inv_slide_dict:
            raise ValueError("Unknown slide pattern {}".format(event.pattern))

        pattern = inv_slide_dict[event.pattern]
        delay = round(event.delay * resolution)
        duration = round(event.duration * resolution)
        result = template.format(
            pattern,
            measure[0],
            measure[1],
            event.position,
            delay,
            duration,
            event.end_position,
        )
    elif isinstance(event, TouchTapNote):
        template = "{}\t{}\t{}\t{}\t{}\t{}\t{}\n"
        name = inv_note_dict[event.note_type.value]
        fireworks = 1 if event.is_firework else 0
        result = template.format(
            name,
            measure[0],
            measure[1],
            event.position,
            event.region,
            fireworks,
            event.size,
        )
    elif isinstance(event, TouchHoldNote):
        template = "{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}\n"
        name = inv_note_dict[event.note_type.value]
        duration = round(event.duration * resolution)
        fireworks = 1 if event.is_firework else 0
        result = template.format(
            name,
            measure[0],
            measure[1],
            event.position,
            duration,
            event.region,
            fireworks,
            event.size,
        )
    else:
        raise TypeError("Unknown note type {}".format(event.note_type.value))

    return result



from typing import Union, Callable, Sequence
import copy

from ..event import MaiNote, NoteType
from ..maisdt import (
    MaiSDT,
    HoldNote as SDTHoldNote,
    SlideStartNote as SDTSlideStartNote,
)
from ..maima2 import MaiMa2, TapNote, HoldNote, SlideNote, TouchTapNote, TouchHoldNote

ma2_slide_dict = {
    "SI_": 1,
    "SCL": 2,
    "SCR": 3,
    "SUL": 4,
    "SUR": 5,
    "SSL": 6,
    "SSR": 7,
    "SV_": 8,
    "SXL": 9,
    "SXR": 10,
    "SLL": 11,
    "SLR": 12,
    "SF_": 13,
}


def _default_touch_converter(
    sdt: MaiSDT, touch_note: Union[TouchTapNote, TouchHoldNote]
) -> None:
    if isinstance(touch_note, TouchTapNote) and touch_note.region == "C":
        sdt.add_tap(measure=touch_note.measure, position=0)
    elif isinstance(touch_note, TouchTapNote):
        sdt.add_tap(measure=touch_note.measure, position=touch_note.position)
    elif isinstance(touch_note, TouchHoldNote) and touch_note.region == "C":
        sdt.add_hold(
            measure=touch_note.measure, position=0, duration=touch_note.duration
        )


def ma2_to_sdt(
    ma2: MaiMa2,
    touch_converter: Callable[
        [MaiSDT, Union[TouchTapNote, TouchHoldNote]], None
    ] = _default_touch_converter,
    convert_touch: bool = False,
) -> MaiSDT:
    sdt = MaiSDT()
    convert_notes(sdt, ma2.notes, touch_converter, convert_touch)
    sdt.notes.sort()

    # Compensate for bpm changes
    event_list = [note for note in sdt.notes + ma2.bpms]

    event_list.sort(key=lambda x: x.measure)
    previous_measure = 0.0
    equivalent_current_measure = 0.0
    equivalent_notes = []
    initial_bpm = ma2.get_bpm(0)
    for event in event_list:
        current_measure = event.measure
        current_bpm = ma2.get_bpm(current_measure)

        # current_bpm_eps takes the bpm slightly before the current measure
        # for compensating gaps. This is because if there is a bpm change at
        # the exact same time as a note, equivalent_gap will be wrongly affected.
        if current_measure > 1.0:
            current_bpm_eps = ma2.get_bpm(current_measure - 0.0001)
        else:
            current_bpm_eps = current_bpm

        gap = current_measure - previous_measure
        scale = initial_bpm / current_bpm
        scale_inf = initial_bpm / current_bpm_eps
        equivalent_gap = gap * scale_inf
        equivalent_current_measure += equivalent_gap

        previous_measure = current_measure
        if not isinstance(event, MaiNote):
            continue

        note = copy.deepcopy(event)
        note.measure = equivalent_current_measure

        if isinstance(note, SDTHoldNote):
            note.duration = note.duration * scale
        elif isinstance(note, SDTSlideStartNote):
            note.duration = note.duration * scale
            note.delay = note.delay * scale

        equivalent_notes.append(note)

    sdt.notes = equivalent_notes
    return sdt


def convert_notes(
    sdt: MaiSDT,
    ma2_notes: Sequence[MaiNote],
    touch_converter: Callable[[MaiSDT, Union[TouchTapNote, TouchHoldNote]], None],
    convert_touch: bool,
) -> None:
    skipped_notes = 0
    for ma2_note in ma2_notes:
        note_type = ma2_note.note_type
        if isinstance(ma2_note, TapNote):
            is_break = note_type in [NoteType.break_tap, NoteType.break_tap]
            is_star = note_type in [NoteType.star, NoteType.break_star]
            sdt.add_tap(
                measure=ma2_note.measure,
                position=ma2_note.position,
                is_break=is_break,
                is_star=is_star,
            )
        elif isinstance(ma2_note, HoldNote):
            # Hold, and ex hold
            sdt.add_hold(
                measure=ma2_note.measure,
                position=ma2_note.position,
                duration=ma2_note.duration,
            )
        elif isinstance(ma2_note, SlideNote):
            # Complete slide
            # SDT slide durations include the delay unlike in ma2
            sdt.add_slide(
                measure=ma2_note.measure,
                start_position=ma2_note.position,
                end_position=ma2_note.end_position,
                duration=ma2_note.duration + ma2_note.delay,
                pattern=ma2_note.pattern,
                delay=ma2_note.delay,
            )
        elif isinstance(ma2_note, (TouchTapNote, TouchHoldNote)):
            # Touch tap, and touch hold
            if convert_touch:
                touch_converter(sdt, ma2_note)
            else:
                skipped_notes += 1
        else:
            print("Warning: Unknown note type {}".format(note_type))

    if skipped_notes > 0:
        print("Skipped {} touch note(s)".format(skipped_notes))



from __future__ import annotations
import re
from typing import Union, List, Dict, Optional

from .sdtnote import TapNote, HoldNote, SlideStartNote, SlideEndNote
from ..event import NoteType
from ..time import second_to_measure


class MaiSDT:
    """A class that represents one sdt chart. Only contains notes,
    and does not include information such as bpm, song name,
    chart difficulty, composer, chart maker, etc.
    It only contains enough information to build a working sdt
    chart file.

    Attributes:
        notes: Contains notes of the chart.
    """

    def __init__(self) -> None:
        self.notes: List[Union[TapNote, HoldNote, SlideStartNote, SlideEndNote]] = []
        self.start_slide_notes: Dict[int, Dict[str, Union[int, float]]] = {}
        self.slide_count = 1

    @classmethod
    def open(cls, path: str, encoding: str = "utf-8") -> MaiSDT:
        sdt = cls()
        with open(path, "r", encoding=encoding) as file:
            for line in file:
                if line in ["\n", "\r\n"]:
                    continue

                if re.search(r"\.srt", path) is None:
                    sdt.parse_line(line)
                else:
                    sdt.parse_srt_line(line)

        return sdt

    def parse_line(self, line: str) -> None:
        """Parse a non-SRT comma-separated line.

        Args:
            line: A non-SRT comma-separated line.

        Raises:
            ValueError: When the number of columns are not between 7 and 9.
            RuntimeError: When an end slide is declared before an beginning slide.
            TypeError: When an unknown note type is given,
        """
        values = line.rstrip().rstrip(",").replace(" ", "").split(",")
        if not (7 <= len(values) <= 9):
            raise ValueError(f"Line has invalid number of columns! {len(values)}")
        measure = float(values[0]) + float(values[1])
        position = int(values[3])
        note_type = int(values[4])

        if note_type in [1, 3, 4, 5]:
            # Regular tap note, break tap note, star note,
            is_star = note_type in [4, 5]
            is_break = note_type in [3, 5]
            self.add_tap(
                measure=measure, position=position, is_break=is_break, is_star=is_star
            )
        elif note_type == 2:
            # Hold note
            duration = float(values[2])
            self.add_hold(measure, position, duration)
        elif note_type == 0:  # Start slide
            if len(values) == 9:
                # SDT includes delay
                delay = float(values[8])
            else:
                delay = 0.25

            duration = float(values[2])
            slide_id = int(values[5])
            pattern = int(values[6])
            start_slide = {
                "position": position,
                "measure": measure,
                "duration": duration,
                "delay": delay,
                "pattern": pattern,
            }
            self.start_slide_notes[slide_id] = start_slide
        elif note_type == 128:  # End slide
            slide_id = int(values[5])

            # Get information about corresponding start slide
            if slide_id not in self.start_slide_notes:
                raise RuntimeError("End slide is declared before slide start!")

            start_slide = self.start_slide_notes[slide_id]
            start_measure = start_slide["measure"]
            start_position = start_slide["position"]
            duration = start_slide["duration"]
            delay = start_slide["delay"]
            slide_pattern = start_slide["pattern"]
            self.add_slide(
                start_measure,
                int(start_position),
                position,
                duration,
                int(slide_pattern),
                delay,
            )
        else:
            raise TypeError("Unknown note type {}".format(note_type))

    def parse_srt_line(self, line: str) -> None:
        """Parse an SRT comma-separated line.

        Args:
            line: An SRT comma-separated line.

        Raises:
            ValueError: When the number of columns are not between 7 and 9.
            RuntimeError: When an end slide is declared before an beginning slide.
            TypeError: When an unknown note type is given,
        """
        srt_slide_to_later_dict = {
            0: 1,
            1: 3,
            2: 2,
        }
        values = line.rstrip().rstrip(",").replace(" ", "").split(",")
        if len(values) != 7:
            raise ValueError(f"SRT should have 7 columns. Given: {len(values)}")

        measure = float(values[0]) + float(values[1])
        position = int(values[3])
        note_type = int(values[4])

        if note_type in [0, 4]:
            # 0: Regular tap note or star and start slide note
            # 4: Break tap note
            slide_id = int(values[5])
            if slide_id != 0:
                # Tap notes with a non-zero slide id are stars and start slide
                self.add_tap(measure=measure, position=position, is_star=True)
                duration = float(values[2])
                # Slide patterns in SZT, and later, starts at 1
                pattern = srt_slide_to_later_dict[int(values[6])]
                start_slide: Dict[str, Union[float, int]] = {
                    "position": position,
                    "measure": measure,
                    "duration": duration,
                    "pattern": pattern,
                }
                self.start_slide_notes[slide_id] = start_slide
            else:
                is_break = note_type == 4
                self.add_tap(measure=measure, position=position, is_break=is_break)
        elif note_type == 2:  # Hold note
            duration = float(values[2])
            self.add_hold(measure, position, duration)
        elif note_type == 128:
            slide_id = int(values[5])
            # Get information about corresponding start slide
            if slide_id not in self.start_slide_notes:
                raise RuntimeError("End slide is declared before slide start!")

            start: Dict[str, Union[float, int]] = self.start_slide_notes[slide_id]
            start_measure = start["measure"]
            start_position = start["position"]
            duration = start["duration"]
            slide_pattern = start["pattern"]
            self.add_slide(
                start_measure,
                int(start_position),
                position,
                duration,
                int(slide_pattern),
            )
        else:
            raise TypeError("Unknown note type {}".format(note_type))

    def add_tap(
        self,
        measure: float,
        position: int,
        is_break: bool = False,
        is_star: bool = False,
    ) -> None:
        """Adds a tap note to the list of notes.

        Args:
            measure: Time when the tap happens, in terms of measures.
            position: Button where the tap note happens.
            is_break: Whether a tap note is a break note.
            is_star: Whether a tap note is a star note.

        Examples:
            Add a regular tap note at measure 1 at button 2,
            and a break tap note at measure 2 at button 7.

            >>> sdt = MaiSDT()
            >>> sdt.add_tap(1, 2)
            >>> sdt.add_tap(2, 7, is_break=True)
        """
        tap_note = TapNote(
            measure=measure, position=position, is_break=is_break, is_star=is_star
        )
        self.notes.append(tap_note)

    def del_tap(self, measure: float, position: int) -> None:
        """Deletes a tap note from the list of notes.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the tap note happens.

        Examples:
            Create a break tap note at measure 26.75 at button 4. Then delete it.

            >>> sdt = MaiSDT()
            >>> sdt.add_tap(26.75, 4, is_break=True)
            >>> sdt.del_tap(26.75, 4)
        """
        tap_notes = [
            x
            for x in self.notes
            if isinstance(x, TapNote)
            and x.measure == measure
            and x.position == position
        ]
        for note in tap_notes:
            self.notes.remove(note)

    def add_hold(self, measure: float, position: int, duration: float) -> None:
        """Adds a hold note to the list of notes.

        Args:
            measure: Time when the hold starts, in terms of measures.
            position: Button where the hold happens.
            duration: Total duration of the hold, in terms of measures.

        Examples:
            Add a regular hold note at button 5 at measure 1.5, with
            duration of 2.75 measures.

            >>> sdt = MaiSDT()
            >>> sdt.add_hold(1.5, 5, 2.75)
        """
        hold_note = HoldNote(measure=measure, position=position, duration=duration)
        self.notes.append(hold_note)

    def del_hold(self, measure: float, position: int) -> None:
        """Deletes the matching hold note in the list of notes. If there are multiple
        matches, all matching notes are deleted. If there are no match, nothing happens.

        Args:
            measure: Time when the note starts, in terms of measures.
            position: Button where the hold note happens.

        Examples:
            Add a regular hold note at button 0 at measure 3.25 with duration of 2 measures
            and delete it.

            >>> sdt = MaiSDT()
            >>> sdt.add_hold(3.25, 0, 2)
            >>> sdt.del_hold(3.25, 0)
        """
        hold_notes = [
            x
            for x in self.notes
            if isinstance(x, HoldNote)
            and x.measure == measure
            and x.position == position
        ]
        for note in hold_notes:
            self.notes.remove(note)

    # Note: SDT slide duration includes delay
    def add_slide(
        self,
        measure: float,
        start_position: int,
        end_position: int,
        duration: float,
        pattern: int,
        delay: float = 0.25,
    ) -> None:
        """Adds both a start slide and end slide note to the list of notes.

        Args:
            measure: Time when the slide starts, in
                terms of measures.
            start_position: Button where the slide starts.
            end_position: Button where the slide ends.
            duration: Total duration of the slide, in terms of
                measures. Includes slide delay.
            pattern: Numerical representation of the slide pattern.
            delay: Duration from when the slide appears and when it
                starts to move, in terms of measures. Defaults to 0.25.

        Examples:
            Add a slide at measure 2 from button 6 to button 3 with
            duration of 1.75 measures, delay of 0.25 measures,
            pattern of 1.

            >>> sdt = MaiSDT()
            >>> sdt.add_slide(2, 6, 3, 1.75, 1)
        """
        slide_id = self.slide_count
        start_slide = SlideStartNote(
            measure=measure,
            position=start_position,
            pattern=pattern,
            duration=duration,
            slide_id=slide_id,
            delay=delay,
        )
        end_measure = measure + duration
        end_slide = SlideEndNote(
            measure=end_measure,
            position=end_position,
            pattern=pattern,
            slide_id=slide_id,
        )
        self.notes.append(start_slide)
        self.notes.append(end_slide)
        self.slide_count += 1

        star_notes = [
            x
            for x in self.notes
            if isinstance(x, TapNote)
            and x.note_type in [NoteType.star, NoteType.break_star]
            and x.measure == measure
            and x.position == start_position
        ]
        for star_note in star_notes:
            star_note.amount += 1

    def del_slide(self, measure: float, start_position: int, end_position: int) -> None:
        start_slides = [
            x
            for x in self.notes
            if isinstance(x, SlideStartNote)
            and x.measure == measure
            and x.position == start_position
        ]
        end_slides: List[SlideEndNote] = []
        for note in start_slides:
            slides = [
                x
                for x in self.notes
                if isinstance(x, SlideEndNote)
                and x.slide_id == note.slide_id
                and x.position == end_position
            ]
            end_slides += slides

        correct_start_slides = []
        for note in end_slides:
            slides = [
                x
                for x in self.notes
                if isinstance(x, SlideStartNote) and x.slide_id == note.slide_id
            ]
            correct_start_slides += slides

        star_notes = [
            x
            for x in self.notes
            if isinstance(x, TapNote)
            and x.note_type in [NoteType.star, NoteType.break_star]
            and x.measure == measure
            and x.position == start_position
        ]

        for note in correct_start_slides:
            self.notes.remove(note)

        for note in end_slides:
            self.notes.remove(note)

        for star_note in star_notes:
            star_note.amount -= 1

    def offset(self, offset: Union[float, str], bpm: Optional[float] = None) -> None:
        if isinstance(offset, float):
            offset_mes = offset
        elif isinstance(offset, str) and offset[-1] in ["s", "S"]:
            if bpm is None:
                raise ValueError("No BPM given")

            offset_mes = second_to_measure(float(offset[:-1]), bpm)
        elif isinstance(offset, str) and "/" in offset:
            fraction = offset.split("/")
            if len(fraction) != 2:
                raise ValueError(f"Invalid fraction: {offset}")

            offset_mes = int(fraction[0]) / int(fraction[1])
        else:
            offset_mes = float(offset)

        for note in self.notes:
            note.measure = round((note.measure + offset_mes) * 10000.0) / 10000.0

    def export(self) -> str:
        """Generates an sdt text from all the notes defined.

        Returns:
            A multiline string. The returned
            string is a complete and functioning sdt text and should
            be stored as-is in a text file with an sdt file extension.
        """
        self.notes.sort()
        result = ""
        for note in self.notes:
            result += str(note)

        return result



import math
from typing import Union

from ..event import MaiNote, NoteType


class TapNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        is_break: bool = False,
        is_star: bool = False,
    ) -> None:
        """Produces an sdt tap note.

        Note:
            Please use MaiSDT class' add_tap method for adding
            tap notes.

        Args:
            measure: Time when the tap happens, in terms of measures.
            position: Button where the tap note happens.
            is_break: Whether a tap note is a break note.
        """
        measure = round(10000.0 * measure) / 10000.0
        if is_break and is_star:
            super().__init__(measure, position, NoteType.break_star)
            self.amount = 0
        elif is_break:
            super().__init__(measure, position, NoteType.break_tap)
        elif is_star:
            super().__init__(measure, position, NoteType.star)
            self.amount = 0
        else:
            super().__init__(measure, position, NoteType.tap)

    def __str__(self) -> str:
        return sdt_note_to_str(self)


class SlideStartNote(MaiNote):
    def __init__(
        self,
        measure: float,
        position: int,
        pattern: int,
        duration: float,
        slide_id: int,
        delay: float = 0.25,
    ) -> None:
        """Produces an sdt slide start note.

        Note:
            Please use MaiSDT class' add_slide method for adding
            slide start and slide end.

        Args:
            measure: Time when the slide starts, in terms of measures.
            position: Button where the slide starts.
            slide_id: Unique non-zero integer for slide start
                and end pair.
            pattern: Numerical representation of the slide pattern.
            duration: Total duration of the slide, in terms of
                measures. Includes slide delay.
            delay: Duration from when the slide appears and when it
                starts to move, in terms of measures.
        """
        if slide_id <= 0:
            raise ValueError("Slide id is not positive " + str(slide_id))
        if pattern <= 0:
            raise ValueError("Slide pattern is not positive " + str(pattern))
        if duration <= 0:
            raise ValueError("Slide duration is not positive " + str(duration))
        if delay < 0:
            raise ValueError("Slide delay is negative " + str(delay))

        measure = round(10000.0 * measure) / 10000.0
        duration = round(10000.0 * duration) / 10000.0
        delay = round(10000.0 * delay) / 10000.0
        super().__init__(measure, position, NoteType.start_slide)
        self.slide_id = slide_id
        self.pattern = pattern
        self.delay = delay
        self.duration = duration

    def __str__(self) -> str:
        return sdt_note_to_str(self)


class SlideEndNote(MaiNote):
    def __init__(
        self, measure: float, position: int, pattern: int, slide_id: int
    ) -> None:
        """Produces an sdt slide end note.

        Note:
            Please use MaiSDT class' add_slide method for adding
            slide start and slide end.

        Args:
            measure: Time when the slide ends, in terms of measures.
            position: Button where the slide ends.
            slide_id: Unique non-zero integer for slide start
                and end pair.
            pattern: Numerical representation of the slide pattern.
        """
        if slide_id <= 0:
            raise ValueError("Slide id is not positive " + str(slide_id))
        if pattern <= 0:
            raise ValueError("Slide pattern is not positive " + str(pattern))

        measure = round(10000 * measure) / 10000
        super().__init__(measure, position, NoteType.end_slide)
        self.slide_id = slide_id
        self.pattern = pattern

    def __str__(self) -> str:
        return sdt_note_to_str(self)


class HoldNote(MaiNote):
    def __init__(self, measure: float, position: int, duration: float) -> None:
        """Produces an sdt hold note.

        Note:
            Please use MaiSDT class' add_hold method for adding
            hold notes.

        Args:
            measure: Time when the hold starts, in terms of measures.
            position: Button where the hold happens.
            duration: Total duration of the hold, in terms of measures.
        """
        if duration < 0:
            raise ValueError(f"Hold duration is negative: {duration}")

        measure = round(10000.0 * measure) / 10000.0
        duration = round(10000.0 * duration) / 10000.0
        super().__init__(measure, position, NoteType.hold)
        self.duration = duration

    def __str__(self) -> str:
        return sdt_note_to_str(self)


def sdt_note_to_str(
    note: Union[TapNote, HoldNote, SlideEndNote, SlideStartNote]
) -> str:
    """Prints note into sdt-compatible lines.

    Args:
        note: A MaiNote to be converted to a sdt string.

    Returns:
        A single line string.
    """
    measure = math.modf(note.measure)
    note_type = note.note_type
    if isinstance(note, (HoldNote, SlideStartNote)):
        note_duration = note.duration
    elif isinstance(note, SlideEndNote):
        note_duration = 0
    else:
        note_duration = 0.0625

    if isinstance(note, (SlideStartNote, SlideEndNote)):
        slide_id = note.slide_id
        pattern = note.pattern
    else:
        slide_id = 0
        pattern = 0

    if isinstance(note, TapNote) and note.note_type in [
        NoteType.star,
        NoteType.break_star,
    ]:
        slide_amount = note.amount
    else:
        slide_amount = 0

    delay = 0.0 if not isinstance(note, SlideStartNote) else note.delay
    line_template = (
        "{:.4f}, {:.4f}, {:.4f}, {:2d}, {:3d}, {:3d}, {:2d}, {:2d}, {:.4f},\n"
    )
    result = line_template.format(
        measure[1],
        measure[0],
        note_duration,
        note.position,
        note_type.value,
        slide_id,
        pattern,
        slide_amount,
        delay,
    )
    return result
